import os, pwd, grp
from time import sleep, time
from shutil import copy, move
# from multiprocessing import Process
import multiprocessing
from subprocess import run, Popen, PIPE
# from zipfile import ZipFile, ZIP_DEFLATED

# lock = f"compiler.lock"

# if os.path.isfile(lock):
#     print("Another instance of the script is already running.")
#     exit(1)

# with open(lock, "w") as f:
    # Write the process ID (PID) to the lock file
    # f.write(str(os.getpid()))

MAX_VRAD_TIME = 30
MAX_VVIS_TIME = 15
MAX_VBSP_TIME = 2

BSP_LIFETIME = 3600 #how long until COMPILED bsp's are cleaned up in seconds


SPARE_THREADS = 0 #will test just letting the linux scheduler + -low parameter do their thing
MAX_THREADS = os.cpu_count() - SPARE_THREADS
# MAX_THREADS = 48

# def get_thread_cpu_utilization():
#     cmd = "top -H -b -n 1"  # Command to get thread CPU utilization
#     result = check_output(cmd, shell=True, universal_newlines=True)
#     print(result)

# i = 0
# while i < 56:
#     get_thread_cpu_utilization()
#     i += 1
# for thread_id, utilization in get_thread_cpu_utilization.items():
    # print(f"Thread ID {thread_id}: CPU Utilization {utilization}%")

def get_file_owner(file_path):
    try:
        # Get the UID (User ID) of the file's owner
        file_stat = os.stat(file_path)
        owner_uid = file_stat.st_uid

        # Get the username associated with the UID
        owner_info = pwd.getpwuid(owner_uid)
        owner_name = owner_info.pw_name

        return owner_name
    except Exception as e:
        return str(e)
    
def change_file_owner(file_path, new_owner):
    try:
        # Use the chown command to change file ownership
        run(["chown", new_owner, file_path])
        return True
    except Exception as e:
        return str(e)

def get_file_group(file_path):
    try:
        # Get the group ID (GID) of the file's group
        group_gid = os.stat(file_path).st_gid

        # Get the group name associated with the GID
        group_info = grp.getgrgid(group_gid)
        group_name = group_info.gr_name

        return group_name
    except Exception as e:
        return str(e)

def set_file_group_permissions(file_path, group, permissions):
    try:
        # Get the current group and owner permissions
        current_mode = os.stat(file_path).st_mode

        # Set the group permissions
        new_mode = current_mode | permissions

        # Apply the new mode to the file
        os.chmod(file_path, new_mode)

        # Change the group ownership (optional)
        os.chown(file_path, -1, os.getgrnam(group).gr_gid)

        return True
    except Exception as e:
        return str(e)
    
def get_file_group_permissions(file_path):
    try:
        # Get the file's stat object
        file_stat = os.stat(file_path)

        # Extract the group permissions part of the file mode
        group_permissions = file_stat.st_mode & 0o70  # Mask to keep only group permissions

        return oct(group_permissions)
    except Exception as e:
        return str(e)
    
def set_file_permissions(file_path, permissions):
    try:
        # Use the os.chmod() function to set the file permissions
        os.chmod(file_path, permissions)
        return True
    except Exception as e:
        return str(e)

def running_on_wine():
    try: 
        if 'WINELOADERNOEXEC' in os.environ or 'wine' in open('/proc/version').read():
            return True
    except:
        return False
    
def remove_filename_from_path(dir):
    return '/'.join(dir.replace('\\', '/').split('/')[:-1])

dir = os.getcwd()
dir = remove_filename_from_path(dir)

mapnames, owners = [], []   
tooldir = '/var/tf2testingserver/bin'
assetdir = '/var/www/html/tf'

try:
    os.symlink(assetdir, f'{remove_filename_from_path(tooldir)}/download')
except Exception as FileExistsError:
    os.remove(f'{remove_filename_from_path(tooldir)}/download')
    os.symlink(assetdir, f'{remove_filename_from_path(tooldir)}/download')
# tooldir = '/var/www/html/bin'
# tooldir = f'{dir}/bin'
# tooldir = r'C:\Program Files (x86)\Steam\steamapps\common\Team Fortress 2\bin'

#configured to grab files from the /tf/maps/ folder and move them to /bin/.
#compiles in /bin/, then moves the result back to /tf/maps/ and deletes the vmf
def run_compiler(tool, threads, addargs, n):

    #all outputs are done through stderr instead of stdout
    os.chdir(tooldir)
    fulltool = ['nice', '-n', "20", 'wine', f'{tooldir}/{tool}']
    # fulltool = ['chrt', '-p', '-i', '0', 'wine', f'{tooldir}/{tool}']
    # os.environ['WINEDEBUG'] = 'trace+all'
    # fulltool = [f'{tooldir}/{tool}']

    params = ['-low', '-threads', str(threads)]

    if tool == 'vbsp.exe':
        params.append('-insert_search_path')
        params.append(assetdir)

    for a in addargs:
        params.append(a)

    params.append('-game')
    params.append(tooldir)
    params.append(n)
    
    Popen(fulltool + params, stderr=PIPE).wait()
    os.chdir(dir)
    # compiler = Popen(f'{fulltool} {addargs} -threads {threads} -game "{remove_filename_from_path(tooldir)}/tf" {n}', stderr=PIPE)

m = 0
def collect_maps():
    global mapnames, m

    n = ''
    initial_size, current_size = -1, 0
    sleeptime = 2
    mod = m % BSP_LIFETIME

    for filename in os.listdir():
        
        if mod == 0 and m >= (BSP_LIFETIME / sleeptime) and filename.split('.')[0].endswith('COMPILED'): 
            os.remove(filename)
            print(f'BSP lifetime exceeded ({BSP_LIFETIME / 60} Minutes)! Deleting {filename}.')

        if (not filename.endswith('.vmf')) or ('instance' in filename or 'skybox' in filename): continue

        #don't compile maps while they're still being uploaded
        initial_size = os.stat(filename).st_size
        sleep(1)
        current_size = os.stat(filename).st_size
        n = filename[:-4]

    if n in mapnames:
        print(f'Waiting for {n} to download...')

    elif initial_size == current_size:
        mapnames.append(n)

    m += 1
    print(f'Waiting for maps x{m}')
    sleep(sleeptime)
    

def compile_maps(threads, addargs = []):
    global m, owners
    for i, n in enumerate(mapnames):
        finalmapdir = f'{dir}/maps/{n}'
        bsp = f'{finalmapdir}.bsp'
        vmf = f'{finalmapdir}.vmf'
        log = f'{finalmapdir}.log'
        zip = f'{finalmapdir}.zip'
        owners.append(None)

        #save owner information
        owners[i] = [get_file_owner(vmf), get_file_group(bsp), get_file_group_permissions(vmf)]

        # [input(o) for o in owners]
        try:
            copy(vmf, f'{tooldir}/{n}.vmf')
        except Exception as FileNotFoundError:
            if not n in mapnames: continue
            mapnames.remove(n)
            continue
        #set default args
        vbsp = multiprocessing.Process(target=run_compiler, args=('vbsp.exe', threads, addargs, n))

        if '-final' in addargs:
            addargs = []

        vvis = multiprocessing.Process(target=run_compiler, args=('vvis.exe', threads, addargs, n))

        if addargs == []:
            addargs =  ['-final', '-StaticPropLighting', '-StaticPropPolys', '-TextureShadows']
        
        #-both has many problems, compile separately instead

        #HDR
        addargs.append('-ldr')
        vrad1 = multiprocessing.Process(target=run_compiler, args=('vrad.exe', threads, addargs, n))

        #LDR
        addargs[-1] == '-hdr'
        vrad2 = multiprocessing.Process(target=run_compiler, args=('vrad.exe', threads, addargs, n))
        
        #TODO add basic file packing, ideally just hook this up to compilepal packing and use our existing fileserver to get assets.
        processes = [[vbsp, MAX_VBSP_TIME, "VBSP"], [vvis, MAX_VVIS_TIME, "VVIS"], [vrad1, MAX_VRAD_TIME, "VRAD (LDR)"], [vrad2, MAX_VRAD_TIME, 'VRAD (HDR)']]
        start = time()
        for p in processes:
            process, timelimit, tool = p[0], p[1], p[2]
            print(f'Starting {tool}...')
            process.start()
            process.join(timelimit*60)
            if tool == 'VVIS': os.remove(f'{tooldir}/{n}.vmf')
            sleep(1)

            if process.is_alive():
                print(f"{tool} has been running for exceedingly long ({timelimit} minutes). Skipping...")
                print(f"{tool} has been running for exceedingly long ({timelimit} minutes). Skipping...")
                print(f"{tool} has been running for exceedingly long ({timelimit} minutes). Skipping...")
                print(f"{tool} has been running for exceedingly long ({timelimit} minutes). Skipping...")
                print(f"{tool} has been running for exceedingly long ({timelimit} minutes). Skipping...")
                process.terminate()
                sleep(1)
                continue
        end = time()
        # os.chdir(dir)
        #move back
        minutes, seconds = int((end - start) / 60), int((end - start) % 60)
        final = f'{minutes} MINUTES, {seconds} SECONDS'

        try:
            os.remove(f'{finalmapdir}.vmf.bz2')
        except Exception as FileNotFoundError:
            return str(FileNotFoundError)

        try:
            os.remove(vmf)
        except Exception as FileNotFoundError:
            return str(FileNotFoundError)
        finally:
            process.terminate()
            
        try:
            bsp = f'{finalmapdir}COMPILED.bsp'
            move(f'{tooldir}/{n}.bsp', bsp)
        
            change_file_owner(bsp, owners[i][0])
            # set_file_group_permissions(bsp, owners[i][1], owners[i][2])
            set_file_group_permissions(bsp, owners[i][0], owners[i][2])
            set_file_permissions(bsp, 0o775)


        except Exception as FileNotFoundError:
            print(FileNotFoundError)
            
        try:
            
            # try:
            #     with ZipFile(zip, 'w', ZIP_DEFLATED) as z:
            #         z.write(log)
            # except Exception as e:
            #     return str(e)
            log = f'{finalmapdir}COMPILED.log'
            with open(log, 'a', encoding='utf-8') as file:
                file.write(log + f'\n || FINAL TIME: {final} ||')
            move(f'{tooldir}/{n}.log', f'{log}.txt')

            change_file_owner(log, owners[i][0])
            # set_file_group_permissions(log, owners[i][1], owners[i][2])
            set_file_group_permissions(log, owners[i][0], owners[i][2])
            set_file_permissions(log, 0o775)
            # move(f'{tooldir}/{n}.zip', zip)


        except Exception as FileNotFoundError:
            print(FileNotFoundError)

        mapnames.remove(n)
        print(f'Compiled {n} in {final}')
        m = 0
        sleep(3)
    owners.clear()

if __name__ == '__main__':
    while True:
        collect_maps()
        compile_maps(MAX_THREADS)